;; Generated package description from phpinspect.el  -*- no-byte-compile: t -*-
(define-package "phpinspect" "1.1.0" "PHP parsing and code intelligence package" '((compat "29")) :commit "de248b090175b1ca2161a33eae963c6f9db5f45d" :url "https://elpa.gnu.org/packages/phpinspect.html" :authors '(("Hugo Thunnissen" . "devel@hugot.nl")) :maintainer '("Hugo Thunnissen" . "devel@hugot.nl") :keywords '("php" "languages" "tools" "convenience"))
